# vim: set ts=4

# Copyright 2022-present Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import contextlib
import difflib
import sys
import xmlrpc
from dataclasses import MISSING, InitVar, asdict, dataclass, field, fields
from pathlib import Path
from typing import Any, Dict, List, Set

from voluptuous import MultipleInvalid

from lavacli import colors, schemas
from lavacli.utils import rt_yaml, safe_yaml


def print_file_diff(src, dst):
    if src is None:
        diffs = difflib.unified_diff([], dst.split("\n"), lineterm="")
    elif dst is None:
        diffs = difflib.unified_diff(src.split("\n"), [], lineterm="")
    else:
        diffs = difflib.unified_diff(src.split("\n"), dst.split("\n"), lineterm="")

    print("    | " + "\n    | ".join(list(diffs)[2:]))


class Base:
    @classmethod
    def new(cls, **kwargs):
        fields_names = [f.name for f in fields(cls)]
        i_kwargs = {}
        v_kwargs = {}
        for k in kwargs:
            if k in fields_names:
                v_kwargs[k] = kwargs[k]
            else:
                i_kwargs[k] = kwargs[k]

        return cls(**v_kwargs)

    def diff(self, data: Dict[str, Any]) -> List[str]:
        return [
            f.name for f in fields(self) if getattr(self, f.name) != data.get(f.name)
        ]


@dataclass(frozen=True, order=True)
class GroupDevicePermission:
    name: str
    group: str

    def dump(self):
        data = {k: v for k, v in asdict(self).items()}
        return data if data else None

    def __repr__(self):
        return f"{self.name:<25}{self.group}"


@dataclass
class Device(Base):
    hostname: str
    device_type: str
    worker: str
    description: str = None
    tags: Set[str] = field(default_factory=set)
    permissions: Set[GroupDevicePermission] = field(default_factory=set)
    health: str = None
    retire: InitVar[bool] = None

    def __post_init__(self, retire):
        self.tags = set(self.tags)
        self.permissions = set([GroupDevicePermission(**p) for p in self.permissions])
        self.retire = retire
        if self.retire is True:
            self.health = "RETIRED"

    def diff(self, data: Dict[str, Any]) -> List[str]:
        data = data.copy()
        data["tags"] = set(data["tags"])
        if data["description"] is None:
            data["description"] = ""
        if self.description is None:
            self.description = ""
        if self.health is None:
            # Do nothing if not defined.
            self.health = data["health"]
        if self.retire is False and data["health"] == "RETIRED":
            # set health to UNKNOWN to un-retire.
            self.health = "UNKNOWN"
        return super().diff(data)

    def dump(self, exclude):
        defaults = {f.name: f.default for f in fields(self) if f.default != MISSING}
        data = {}
        for k, v in asdict(self).items():
            if k in defaults and v == defaults[k]:
                continue
            data[k] = v

        if "description" in data and data["description"] in ["", None]:
            del data["description"]
        if not data["tags"]:
            del data["tags"]
        else:
            data["tags"] = sorted(data["tags"])
        del data["hostname"]
        if not data["permissions"] or "permissions" in exclude or "groups" in exclude:
            del data["permissions"]
        else:
            data["permissions"] = [p.dump() for p in sorted(data["permissions"])]
        # Add `retire: true` for retired device.
        if data["health"] == "Retired":
            data["retire"] = True
        del data["health"]
        return data if data else None

    def get_dict(self, base):
        with contextlib.suppress(FileNotFoundError):
            return (base / "devices" / f"{self.hostname}.jinja2").read_text(
                encoding="utf-8"
            )
        return None

    def set_dict(self, base, text):
        (base / "devices").mkdir(parents=True, exist_ok=True)
        with contextlib.suppress(FileNotFoundError):
            return (base / "devices" / f"{self.hostname}.jinja2").write_text(
                text, encoding="utf-8"
            )


@dataclass
class DeviceType(Base):
    name: str
    description: str = ""
    health_disabled: bool = False
    health_denominator: str = "hours"
    health_frequency: int = 24
    aliases: Set[str] = field(default_factory=set)
    display: bool = True
    permissions: Set[GroupDevicePermission] = field(default_factory=set)

    def __post_init__(self):
        self.aliases = set(self.aliases)
        self.permissions = set([GroupDevicePermission(**p) for p in self.permissions])

    def diff(self, data: Dict[str, Any]) -> List[str]:
        data = data.copy()
        data["aliases"] = set(data["aliases"])
        if data["description"] is None:
            data["description"] = ""
        if self.description is None:
            self.description = ""
        return super().diff(data)

    def dump(self, exclude):
        defaults = {f.name: f.default for f in fields(self) if f.default != MISSING}
        data = {}
        for k, v in asdict(self).items():
            if k in defaults and v == defaults[k]:
                continue
            data[k] = v

        if not data["aliases"]:
            del data["aliases"]
        else:
            data["aliases"] = sorted(data["aliases"])
        if "description" in data and data["description"] in ["", None]:
            del data["description"]
        del data["name"]
        if not data["permissions"] or "permissions" in exclude or "groups" in exclude:
            del data["permissions"]
        else:
            data["permissions"] = [p.dump() for p in sorted(data["permissions"])]
        return data if data else None

    def get_health_check(self, base):
        with contextlib.suppress(FileNotFoundError):
            return (base / "health-checks" / f"{self.name}.yaml").read_text(
                encoding="utf-8"
            )
        return None

    def set_health_check(self, base, text):
        (base / "health-checks").mkdir(parents=True, exist_ok=True)
        (base / "health-checks" / f"{self.name}.yaml").write_text(
            text, encoding="utf-8"
        )

    def get_template(self, base):
        with contextlib.suppress(FileNotFoundError):
            return (base / "device-types" / f"{self.name}.jinja2").read_text(
                encoding="utf-8"
            )

    def set_template(self, base, text):
        (base / "device-types").mkdir(parents=True, exist_ok=True)
        (base / "device-types" / f"{self.name}.jinja2").write_text(
            text, encoding="utf-8"
        )


@dataclass(frozen=True, order=True)
class Permission:
    app: str
    model: str
    codename: str

    @classmethod
    def from_str(cls, s):
        app, model, codename = s.split(".")
        return cls(app=app, model=model, codename=codename)

    def dump(self):
        return str(self)

    def __repr__(self):
        return f"{self.app}.{self.model}.{self.codename}"


@dataclass
class Group(Base):
    name: str
    permissions: Set[Permission] = field(default_factory=set)

    def __post_init__(self):
        self.permissions = set([Permission.from_str(p) for p in self.permissions])

    def dump(self, exclude):
        defaults = {f.name: f.default for f in fields(self) if f.default != MISSING}
        data = {}
        for k, v in asdict(self).items():
            if k in defaults and v == defaults[k]:
                continue
            data[k] = v

        if not data["permissions"] or "permissions" in exclude:
            del data["permissions"]
        else:
            data["permissions"] = [p.dump() for p in sorted(data["permissions"])]
        del data["name"]
        return data if data else None


@dataclass
class User(Base):
    username: str
    last_name: str = ""
    first_name: str = ""
    email: str = ""
    is_superuser: bool = False
    is_staff: bool = False
    is_active: bool = True
    ldap: bool = False
    groups: Set[str] = field(default_factory=set)
    permissions: Set[Permission] = field(default_factory=set)

    def __post_init__(self):
        self.groups = set(self.groups)
        self.permissions = set([Permission.from_str(p) for p in self.permissions])

    def diff(self, data: Dict[str, Any]) -> List[str]:
        data = data.copy()
        data["groups"] = set(data["groups"])
        diff = super().diff(data)
        # ldap is user add only.
        if "ldap" in diff:
            diff.remove("ldap")
        # Don't update first_name, last_name and email for ldap user.
        if self.ldap:
            diff = [d for d in diff if d not in ["first_name", "last_name", "email"]]

        return diff

    def dump(self, exclude):
        defaults = {f.name: f.default for f in fields(self) if f.default != MISSING}
        data = {}
        for k, v in asdict(self).items():
            if k in defaults and v == defaults[k]:
                continue
            data[k] = v

        if not data["groups"] or "groups" in exclude:
            del data["groups"]
        else:
            data["groups"] = sorted(data["groups"])
        if not data["permissions"] or "permissions" in exclude:
            del data["permissions"]
        else:
            data["permissions"] = [p.dump() for p in sorted(data["permissions"])]
        del data["username"]
        return data if data else None


@dataclass
class Worker(Base):
    hostname: str
    description: str = ""
    job_limit: int = 0
    health: str = None
    retire: InitVar[bool] = False

    def __post_init__(self, retire):
        self.retire = retire
        if self.retire is True:
            self.health = "RETIRED"

    def diff(self, data: Dict[str, Any]) -> List[str]:
        data = data.copy()
        if data["description"] is None:
            data["description"] = ""
        if self.description is None:
            self.description = ""
        if self.health is None:
            # Do nothing if not defined.
            self.health = data["health"]
        if self.retire is False and data["health"] == "RETIRED":
            # Set health to ACTIVE to un-retire.
            self.health = "ACTIVE"
        return super().diff(data)

    def dump(self):
        defaults = {f.name: f.default for f in fields(self) if f.default != MISSING}
        data = {}
        for k, v in asdict(self).items():
            if k in defaults and v == defaults[k]:
                continue
            data[k] = v

        if "description" in data and data["description"] in ["", None]:
            del data["description"]
        del data["hostname"]
        # Add `retire: true` for retired worker.
        if data["health"] == "Retired":
            data["retire"] = True
        del data["health"]
        return data if data else None

    def get_config(self, base):
        with contextlib.suppress(FileNotFoundError):
            return (base / "workers" / self.hostname / "dispatcher.yaml").read_text(
                encoding="utf-8"
            )

    def set_config(self, base, text):
        (base / "workers" / self.hostname).mkdir(parents=True, exist_ok=True)
        (base / "workers" / self.hostname / "dispatcher.yaml").write_text(
            text, encoding="utf-8"
        )

    def get_env(self, base):
        with contextlib.suppress(FileNotFoundError):
            return (base / "workers" / self.hostname / "env.yaml").read_text(
                encoding="utf-8"
            )

    def set_env(self, base, text):
        (base / "workers" / self.hostname).mkdir(parents=True, exist_ok=True)
        (base / "workers" / self.hostname / "env.yaml").write_text(
            text, encoding="utf-8"
        )

    def get_env_dut(self, base):
        with contextlib.suppress(FileNotFoundError):
            return (base / "workers" / self.hostname / "env-dut.yaml").read_text(
                encoding="utf-8"
            )

    def set_env_dut(self, base, text):
        (base / "workers" / self.hostname).mkdir(parents=True, exist_ok=True)
        (base / "workers" / self.hostname / "env-dut.yaml").write_text(
            text, encoding="utf-8"
        )


@dataclass
class Config:
    device_types: Dict[str, DeviceType]
    devices: Dict[str, Device]
    workers: Dict[str, Worker]
    groups: Dict[str, Group] = field(default_factory=dict)
    users: Dict[str, User] = field(default_factory=dict)

    def __post_init__(self):
        self.devices = {n: Device(hostname=n, **d) for n, d in self.devices.items()}
        self.device_types = {
            n: DeviceType(name=n, **(dt if dt is not None else {}))
            for n, dt in self.device_types.items()
        }
        self.groups = {
            n: Group(name=n, **(grp if grp is not None else {}))
            for n, grp in self.groups.items()
        }
        self.users = {
            n: User(username=n, **(user if user is not None else {}))
            for n, user in self.users.items()
        }
        self.workers = {
            h: Worker(hostname=h, **(w if w is not None else {}))
            for h, w in self.workers.items()
        }

    def _groups(self, exclude):
        return {k: self.groups[k].dump(exclude) for k in self.groups}

    def _users(self, exclude):
        return {k: self.users[k].dump(exclude) for k in self.users}

    def _device_types(self, exclude):
        return {k: self.device_types[k].dump(exclude) for k in self.device_types}

    def _workers(self):
        return {k: self.workers[k].dump() for k in self.workers}

    def _devices(self, exclude):
        return {k: self.devices[k].dump(exclude) for k in self.devices}

    def dump(self, resources, exclude):
        data = {}
        if "device-types" in resources:
            data["device_types"] = self._device_types(exclude)
        if "devices" in resources:
            data["devices"] = self._devices(exclude)
        if "workers" in resources:
            data["workers"] = self._workers()
        if "groups" in resources and "groups" not in exclude:
            data["groups"] = self._groups(exclude)
        if "users" in resources and "users" not in exclude:
            data["users"] = self._users(exclude)

        return data

    def dump_device_types(self, exclude):
        return {"device_types": self._device_types(exclude)}

    def dump_devices(self, exclude):
        return {"devices": self._devices(exclude)}

    def dump_groups(self, exclude):
        return {"groups": self._groups(exclude)}

    def dump_users(self, exclude):
        return {"users": self._users(exclude)}

    def dump_workers(self):
        return {"workers": self._workers()}


class ConfigFile:
    def __init__(self, filepath):
        self.filepath = filepath

    @property
    def filepath(self):
        return self._filepath

    @filepath.setter
    def filepath(self, value):
        if not isinstance(value, Path):
            raise ValueError("filepath must be a pathlib.Path object!")
        self._filepath = value.resolve()

    def is_yaml_file(self):
        with contextlib.suppress(AttributeError):
            if self.filepath.suffix == ".yaml":
                return True
        return False

    def is_dir(self):
        """Treat non yaml file as a directory"""
        if self.is_yaml_file():
            return False
        return True

    @property
    def base_dir(self):
        base_dir = self.filepath
        if self.is_yaml_file():
            base_dir = (self.filepath / ".." / self.filepath.stem).resolve()
        return base_dir

    def write(self, filename, data):
        print(f"{colors.yellow}> {filename}{colors.reset}")
        with filename.open("w") as f:
            rt_yaml.dump(data, f)

    def load(self):
        if self.is_yaml_file():
            return safe_yaml.load(self.filepath.read_text(encoding="utf-8"))
        if self.is_dir():
            data = {}
            files = [*sorted(self.filepath.glob("*.yaml"))]
            for _file in files:
                _config = safe_yaml.load(_file.read_text(encoding="utf-8"))
                data.update(_config)
            return data
        return {}

    def dump(self, lab, resources=[], exclude=[]):
        if self.is_yaml_file():
            self.write(self.filepath, lab.dump(resources, exclude))
        if self.is_dir():
            if "groups" in resources and "groups" not in exclude:
                self.write((self.base_dir / "groups.yaml"), lab.dump_groups(exclude))
            if "users" in resources and "users" not in exclude:
                self.write((self.base_dir / "users.yaml"), lab.dump_users(exclude))
            if "device-types" in resources:
                self.write(
                    (self.base_dir / "device-types.yaml"),
                    lab.dump_device_types(exclude),
                )
            if "devices" in resources:
                self.write((self.base_dir / "devices.yaml"), lab.dump_devices(exclude))
            if "workers" in resources:
                self.write((self.base_dir / "workers.yaml"), lab.dump_workers())


def configure_parser(parser, version):
    sub = parser.add_subparsers(dest="sub_sub_command", help="Sub commands")
    sub.required = True

    if version < (2022, 4):
        return

    # "apply"
    lab_apply = sub.add_parser("apply", help="apply configuration")
    lab_apply.add_argument(
        "--dry-run",
        action="store_true",
        default=False,
        help="Do not update the configuration",
    )
    lab_apply.add_argument(
        "--resources",
        default=[],
        action="append",
        choices=["groups", "users", "devices", "device-types", "workers"],
        help="resources to sync",
    )
    lab_apply.add_argument(
        "--delete",
        action="store_true",
        default=False,
        help="Delete resource that not found in config file",
    )
    lab_apply.add_argument(
        "config",
        type=Path,
        help="Path to a single config file using '.yaml' extension or a directory containing a set of config files",
    )

    # "import"
    lab_import = sub.add_parser("import", help="import configuration")
    lab_import.add_argument(
        "--resources",
        default=[],
        action="append",
        choices=["groups", "users", "devices", "device-types", "workers"],
        help="resources to import",
    )
    lab_import.add_argument(
        "--exclude",
        default=[],
        action="append",
        choices=["groups", "users", "permissions"],
        help="resource to exclude",
    )
    lab_import.add_argument(
        "config",
        type=Path,
        help="Path to a single config file using '.yaml' extension or a directory containing a set of config files",
    )

    # validate
    lab_validate = sub.add_parser("validate", help="validate configuration")
    lab_validate.add_argument("config", type=Path, help="configuration file")


def help_string():
    return "manage lab configuration"


def validate(data: dict) -> bool:
    try:
        schemas.config_schema(data)
        return True
    except MultipleInvalid as exc:
        print(str(exc))
        print(f"{colors.red}Config invalid!{colors.reset}")
        return False


def handle_apply(proxy, options, config):
    if not options.resources:
        options.resources = ["devices", "device-types", "groups", "users", "workers"]

    config_file = ConfigFile(options.config)
    data = config_file.load()
    if not validate(data):
        return 1
    lab = Config(**data)
    base = config_file.base_dir

    print(f"{colors.cyan}> groups{colors.reset}")
    if "groups" not in options.resources:
        print(f"  {colors.yellow}-> SKIP{colors.reset}")
    elif config["version"] >= (2023, 3):
        groups = proxy.auth.groups.list()
        for group in lab.groups.values():
            if group.name in groups:
                print(f"  {colors.green}* {group.name}{colors.reset}")
            else:
                print(f"  {colors.yellow}* {group.name}{colors.reset}")
                if not options.dry_run:
                    proxy.auth.groups.add(group.name)

            data = proxy.auth.groups.show(group.name)
            data["permissions"] = set(
                [Permission.from_str(p) for p in data["permissions"]]
            )
            diff = group.diff(data)
            if "permissions" in diff:
                print(f"    {colors.yellow}-> permissions{colors.reset}")
                missing = group.permissions.difference(set(data["permissions"]))
                for perm in missing:
                    print(f"      {colors.green}+ {perm}{colors.reset}")
                    if not options.dry_run:
                        proxy.auth.groups.perms.add(
                            group.name, perm.app, perm.model, perm.codename
                        )
                missing = set(data["permissions"]).difference(group.permissions)
                for perm in missing:
                    print(f"      {colors.red}- {perm}{colors.reset}")
                    if not options.dry_run:
                        proxy.auth.groups.perms.delete(
                            group.name, perm.app, perm.model, perm.codename
                        )

        if options.delete:
            for group in groups:
                if group not in lab.groups:
                    print(f"  {colors.red}* {group}{colors.reset}")
                    if not options.dry_run:
                        proxy.auth.groups.delete(group)

    print(f"{colors.cyan}> users{colors.reset}")
    if "users" not in options.resources:
        print(f"  {colors.yellow}-> SKIP{colors.reset}")
    elif config["version"] >= (2023, 3):
        users = [user["username"] for user in proxy.auth.users.list()]
        for user in lab.users.values():
            if user.username in users:
                print(f"  {colors.green}* {user.username}{colors.reset}")
            else:
                print(f"  {colors.yellow}* {user.username}{colors.reset}")
                if not options.dry_run:
                    proxy.auth.users.add(
                        user.username,
                        user.first_name or None,
                        user.last_name or None,
                        user.email or None,
                        user.is_active,
                        user.is_staff,
                        user.is_superuser,
                        user.ldap,
                    )

            data = proxy.auth.users.show(user.username)
            data["permissions"] = set(
                [Permission.from_str(p) for p in data["permissions"]]
            )
            diff = user.diff(data)
            update_diff = [n for n in diff if n not in ["groups", "permissions"]]
            if update_diff:
                for name in update_diff:
                    print(
                        f"    {colors.yellow}-> {name}: '{data[name]}' => '{getattr(user, name)}'{colors.reset}"
                    )
                if not options.dry_run:
                    proxy.auth.users.update(
                        user.username,
                        user.first_name if "first_name" in diff else None,
                        user.last_name if "last_name" in diff else None,
                        user.email if "email" in diff else None,
                        user.is_active if "is_active" in diff else None,
                        user.is_staff if "is_staff" in diff else None,
                        user.is_superuser if "is_superuser" in diff else None,
                    )
            if "groups" in diff:
                print(f"    {colors.yellow}-> groups{colors.reset}")
                missing = user.groups.difference(set(data["groups"]))
                for grp in missing:
                    print(f"      {colors.green}+ {grp}{colors.reset}")
                    if not options.dry_run:
                        proxy.auth.users.groups.add(user.username, grp)
                missing = set(data["groups"]).difference(user.groups)
                for grp in missing:
                    print(f"      {colors.red}- {grp}{colors.reset}")
                    if not options.dry_run:
                        proxy.auth.users.groups.delete(user.username, grp)

            if "permissions" in diff:
                print(f"    {colors.yellow}-> permissions{colors.reset}")
                missing = user.permissions.difference(set(data["permissions"]))
                for perm in missing:
                    print(f"      {colors.green}+ {perm}{colors.reset}")
                    if not options.dry_run:
                        proxy.auth.users.perms.add(
                            user.username, perm.app, perm.model, perm.codename
                        )
                missing = set(data["permissions"]).difference(user.permissions)
                for perm in missing:
                    print(f"      {colors.red}- {perm}{colors.reset}")
                    if not options.dry_run:
                        proxy.auth.users.perms.delete(
                            user.username, perm.app, perm.model, perm.codename
                        )

        if options.delete:
            for user in users:
                if user not in lab.users:
                    print(f"  {colors.red}* {user}{colors.reset}")
                    if not options.dry_run:
                        proxy.auth.users.delete(user)

    print(f"{colors.cyan}> device-types{colors.reset}")
    if "device-types" not in options.resources:
        print(f"  {colors.yellow}-> SKIP{colors.reset}")
    else:
        device_types = [dt["name"] for dt in proxy.scheduler.device_types.list(False)]
        for dt in lab.device_types.values():
            if dt.name in device_types:
                print(f"  {colors.green}* {dt.name}{colors.reset}")
            else:
                print(f"  {colors.yellow}* {dt.name}{colors.reset}")
                if not options.dry_run:
                    proxy.scheduler.device_types.add(
                        dt.name,
                        dt.description or None,
                        dt.display or True,
                        # owners_only is deprecated.
                        None,
                        dt.health_frequency or 24,
                        dt.health_denominator or "hours",
                    )
            data = proxy.scheduler.device_types.show(dt.name)
            if config["version"] >= (2023, 3):
                data["permissions"] = set(
                    [GroupDevicePermission(**p) for p in data.get("permissions", [])]
                )
            diff = dt.diff(data)
            dt_diff = [n for n in diff if n not in ["aliases", "permissions"]]
            if dt_diff:
                for name in dt_diff:
                    print(
                        f"    {colors.yellow}-> {name}: '{data[name]}' => '{getattr(dt, name)}'{colors.reset}"
                    )
                if not options.dry_run:
                    proxy.scheduler.device_types.update(
                        dt.name,
                        dt.description if "description" in diff else None,
                        dt.display if "display" in diff else None,
                        None,
                        dt.health_frequency if "health_frequency" in diff else None,
                        dt.health_denominator if "health_denominator" in diff else None,
                        dt.health_disabled if "health_disabled" in diff else None,
                    )
            if "aliases" in diff:
                print(f"    {colors.yellow}-> aliases{colors.reset}")
                missing = dt.aliases.difference(set(data["aliases"]))
                for alias in missing:
                    print(f"      {colors.green}+ {alias}{colors.reset}")
                    if not options.dry_run:
                        proxy.scheduler.device_types.aliases.add(dt.name, alias)
                missing = set(data["aliases"]).difference(dt.aliases)
                for alias in missing:
                    print(f"      {colors.red}- {alias}{colors.reset}")
                    if not options.dry_run:
                        proxy.scheduler.device_types.aliases.delete(dt.name, alias)

            if config["version"] >= (2023, 3) and "permissions" in diff:
                if "groups" not in options.resources:
                    print(f"    {colors.yellow}->  SKIP permissions{colors.reset}")
                else:
                    print(
                        f"    {colors.yellow}->  {'permission':<25}group{colors.reset}"
                    )
                    missing = sorted(
                        dt.permissions.difference(set(data["permissions"]))
                    )
                    for perm in missing:
                        print(f"      {colors.green}+ {perm}{colors.reset}")
                        if not options.dry_run:
                            proxy.scheduler.device_types.perms_add(
                                dt.name, perm.group, perm.name
                            )
                    missing = sorted(
                        set(data["permissions"]).difference(dt.permissions)
                    )
                    for perm in missing:
                        print(f"      {colors.red}- {perm}{colors.reset}")
                        if not options.dry_run:
                            proxy.scheduler.device_types.perms_delete(
                                dt.name, perm.group, perm.name
                            )

            try:
                hc = str(proxy.scheduler.device_types.get_health_check(dt.name))
            except xmlrpc.client.Fault as exc:
                if exc.faultCode != 404:
                    raise
                hc = None
            if dt.get_health_check(base) != hc:
                print(f"    {colors.yellow}-> health-check{colors.reset}")
                print_file_diff(hc, dt.get_health_check(base))
                if not options.dry_run:
                    proxy.scheduler.device_types.set_health_check(
                        dt.name, dt.get_health_check(base)
                    )

            if not data["default_template"] or dt.get_template(base) is not None:
                try:
                    template = str(proxy.scheduler.device_types.get_template(dt.name))
                except xmlrpc.client.Fault as exc:
                    if exc.faultCode != 404:
                        raise
                    template = None
                if dt.get_template(base) != template:
                    print(f"    {colors.yellow}-> template{colors.reset}")
                    print_file_diff(template, dt.get_template(base))
                    if not options.dry_run:
                        proxy.scheduler.device_types.set_template(
                            dt.name, dt.get_template(base)
                        )

        if options.delete:
            for dt in device_types:
                if dt not in lab.device_types:
                    print(f"  {colors.red}* {dt}{colors.reset}")
                    if not options.dry_run:
                        # Hide the device type in the GUI.
                        proxy.scheduler.device_types.update(
                            dt,
                            None,
                            False,
                            None,
                            None,
                            None,
                            None,
                        )

    print(f"{colors.cyan}> workers{colors.reset}")
    if "workers" not in options.resources:
        print(f"  {colors.yellow}-> SKIP{colors.reset}")
    else:
        if config["version"] >= (2023, 3):
            workers = proxy.scheduler.workers.list(True)
        else:
            workers = proxy.scheduler.workers.list()
        for worker in lab.workers.values():
            if worker.hostname in workers:
                print(f"  {colors.green}* {worker.hostname}{colors.reset}")
            else:
                print(f"  {colors.yellow}* {worker.hostname}{colors.reset}")
                if not options.dry_run:
                    proxy.scheduler.workers.add(
                        worker.hostname, worker.description, False
                    )
            data = proxy.scheduler.workers.show(worker.hostname)
            # Active/Maintenance/Retired -> ACTIVE/MAINTENANCE/RETRIED for comparison.
            data["health"] = data["health"].upper()
            diff = worker.diff(data)
            for name in diff:
                print(
                    f"    {colors.yellow}-> {name}: '{data[name]}' => '{getattr(worker, name)}'{colors.reset}"
                )
            if diff and not options.dry_run:
                proxy.scheduler.workers.update(
                    worker.hostname,
                    worker.description if "description" in diff else None,
                    worker.health if "health" in diff else None,
                    worker.job_limit if "job_limit" in diff else None,
                )

            if not data["default_config"] or worker.get_config(base) is not None:
                try:
                    wconfig = str(proxy.scheduler.workers.get_config(worker.hostname))
                except xmlrpc.client.Fault as exc:
                    if exc.faultCode != 404:
                        raise
                    wconfig = None
                if worker.get_config(base) != wconfig:
                    print(f"    {colors.yellow}-> config{colors.reset}")
                    print_file_diff(wconfig, worker.get_config(base))
                    if not options.dry_run:
                        proxy.scheduler.workers.set_config(
                            worker.hostname, worker.get_config(base)
                        )

            if not data["default_env"] or worker.get_env(base) is not None:
                try:
                    wenv = str(proxy.scheduler.workers.get_env(worker.hostname))
                except xmlrpc.client.Fault as exc:
                    if exc.faultCode != 404:
                        raise
                    wenv = None
                if worker.get_env(base) != wenv:
                    print(f"    {colors.yellow}-> env{colors.reset}")
                    print_file_diff(wenv, worker.get_env(base))
                    if not options.dry_run:
                        proxy.scheduler.workers.set_env(
                            worker.hostname, worker.get_env(base)
                        )

            if not data["default_env_dut"] or worker.get_env_dut(base) is not None:
                try:
                    wenv_dut = str(proxy.scheduler.workers.get_env_dut(worker.hostname))
                except xmlrpc.client.Fault as exc:
                    if exc.faultCode != 404:
                        raise
                    wenv_dut = None
                if worker.get_env_dut(base) != wenv_dut:
                    print(f"    {colors.yellow}-> env-dut{colors.reset}")
                    print_file_diff(wenv_dut, worker.get_env_dut(base))
                    if not options.dry_run:
                        proxy.scheduler.workers.set_env_dut(
                            worker.hostname, worker.get_env_dut(base)
                        )

        if options.delete and config["version"] >= (2023, 5):
            for worker in workers:
                if worker not in lab.workers:
                    print(f"  {colors.red}* {worker}{colors.reset}")
                    if not options.dry_run:
                        proxy.scheduler.workers.delete(worker)

    print(f"{colors.cyan}> devices{colors.reset}")
    if "devices" not in options.resources:
        print(f"  {colors.yellow}-> SKIP{colors.reset}")
    else:
        devices = [d["hostname"] for d in proxy.scheduler.devices.list(True)]
        for device in lab.devices.values():
            if device.hostname in devices:
                print(f"  {colors.green}* {device.hostname}{colors.reset}")
            else:
                print(f"  {colors.yellow}* {device.hostname}{colors.reset}")
                if not options.dry_run:
                    proxy.scheduler.devices.add(
                        device.hostname,
                        device.device_type,
                        device.worker,
                        None,
                        None,
                        None,
                        None,
                        device.description,
                    )
            data = proxy.scheduler.devices.show(device.hostname)
            # Good/Unknow/Looping/Bad/Maintenance/Retired ->
            # GOOD/UNKNOWN/LOOPING/BAD/MAINTENANCE/RETIRED for comparison
            data["health"] = data["health"].upper()
            if config["version"] >= (2023, 3):
                data["permissions"] = set(
                    [GroupDevicePermission(**p) for p in data.get("permissions", [])]
                )
            diff = device.diff(data)
            device_diff = [n for n in diff if n not in ["tags", "permissions"]]
            if device_diff:
                for name in device_diff:
                    print(
                        f"    {colors.yellow}-> {name}: '{data[name]}' => '{getattr(device, name)}'{colors.reset}"
                    )
                if not options.dry_run:
                    proxy.scheduler.devices.update(
                        device.hostname,
                        device.worker if "worker" in diff else None,
                        None,
                        None,
                        None,
                        device.health if "health" in diff else None,
                        device.description if "description" in diff else None,
                        device.device_type if "device_type" in diff else None,
                    )
            if "tags" in diff:
                print(f"    {colors.yellow}-> tags{colors.reset}")
                missing = device.tags.difference(set(data["tags"]))
                for tag in missing:
                    print(f"      {colors.green}+ {tag}{colors.reset}")
                    if not options.dry_run:
                        proxy.scheduler.devices.tags.add(device.hostname, tag)
                missing = set(data["tags"]).difference(device.tags)
                for tag in missing:
                    print(f"      {colors.red}- {tag}{colors.reset}")
                    if not options.dry_run:
                        proxy.scheduler.devices.tags.delete(device.hostname, tag)

            if config["version"] >= (2023, 3) and "permissions" in diff:
                if "groups" not in options.resources:
                    print(f"    {colors.yellow}->  SKIP permissions{colors.reset}")
                else:
                    print(
                        f"    {colors.yellow}->  {'permission':<25}group{colors.reset}"
                    )
                    missing = sorted(
                        device.permissions.difference(set(data["permissions"]))
                    )
                    for perm in missing:
                        print(f"      {colors.green}+ {perm}{colors.reset}")
                        if not options.dry_run:
                            proxy.scheduler.devices.perms_add(
                                device.hostname, perm.group, perm.name
                            )
                    missing = sorted(
                        set(data["permissions"]).difference(device.permissions)
                    )
                    for perm in missing:
                        print(f"      {colors.red}- {perm}{colors.reset}")
                        if not options.dry_run:
                            proxy.scheduler.devices.perms_delete(
                                device.hostname, perm.group, perm.name
                            )

            try:
                ddict = str(proxy.scheduler.devices.get_dictionary(device.hostname))
            except xmlrpc.client.Fault as exc:
                if exc.faultCode != 404:
                    raise
                ddict = None

            if device.get_dict(base) != ddict:
                print(f"    {colors.yellow}-> dictionary{colors.reset}")
                print_file_diff(ddict, device.get_dict(base))
                if not options.dry_run:
                    proxy.scheduler.devices.set_dictionary(
                        device.hostname, device.get_dict(base)
                    )

        if options.delete and config["version"] >= (2023, 5):
            for device in devices:
                if device not in lab.devices:
                    print(f"  {colors.red}* {device}{colors.reset}")
                    if not options.dry_run:
                        proxy.scheduler.devices.delete(device)

    return 0


def handle_import(proxy, options, config):
    lab = Config({}, {}, {}, {}, {})

    config_file = ConfigFile(options.config)
    base = config_file.base_dir
    base.mkdir(parents=True, exist_ok=True)

    resources = options.resources
    if resources:
        if "devices" in resources:
            if (
                "device-types" not in resources
                or "workers" not in resources
                or "groups" not in resources
            ):
                sys.stderr.write(
                    "'--resources devices' needs '--resources device-types --resources workers --resources groups'\n"
                )
                return 1
        if "device-types" in resources:
            if "groups" not in resources:
                sys.stderr.write(
                    "'--resources device-types' needs '--resources groups'\n"
                )
                return 1
        if "users" in resources:
            if "groups" not in resources:
                sys.stderr.write("'--resources users' needs '--resources groups'\n")
                return 1
    else:
        resources = ["devices", "device-types", "groups", "users", "workers"]

    excluded_resources = options.exclude

    if config["version"] >= (2023, 3):
        print(f"{colors.cyan}> groups{colors.reset}")
        if "groups" not in resources:
            print(f"  {colors.yellow}-> SKIP{colors.reset}")
        else:
            if "groups" in excluded_resources:
                print(f"  {colors.yellow}-> SKIP{colors.reset}")
            else:
                groups = [grp for grp in proxy.auth.groups.list()]
                for grp in groups:
                    print(f"  {colors.green}* {grp}{colors.reset}")
                    data = proxy.auth.groups.show(grp)
                    lab.groups[grp] = Group.new(
                        name=data["name"], permissions=data["permissions"]
                    )

        print(f"{colors.cyan}> users{colors.reset}")
        if "users" not in resources:
            print(f"  {colors.yellow}-> SKIP{colors.reset}")
        else:
            if "users" in excluded_resources:
                print(f"  {colors.yellow}-> SKIP{colors.reset}")
            else:
                users = [user["username"] for user in proxy.auth.users.list()]
                for user in users:
                    print(f"  {colors.green}* {user}{colors.reset}")
                    data = proxy.auth.users.show(user)
                    lab.users[user] = User.new(**data)

    print(f"{colors.cyan}> device-types{colors.reset}")
    if "device-types" not in resources:
        print(f"  {colors.yellow}-> SKIP{colors.reset}")
    else:
        device_types = [dt["name"] for dt in proxy.scheduler.device_types.list(False)]
        for dt in device_types:
            print(f"  {colors.green}* {dt}{colors.reset}")
            data = proxy.scheduler.device_types.show(dt)
            lab.device_types[dt] = DeviceType.new(**data)

            try:
                hc = str(proxy.scheduler.device_types.get_health_check(dt))
            except xmlrpc.client.Fault as exc:
                if exc.faultCode != 404:
                    raise
                hc = None
            if hc is not None:
                print(f"  {colors.green}  -> health-check{colors.reset}")
                lab.device_types[dt].set_health_check(base, hc)

            if data["default_template"]:
                print(f"  {colors.green}  -> default template{colors.reset}")
            else:
                try:
                    template = str(proxy.scheduler.device_types.get_template(dt))
                except xmlrpc.client.Fault as exc:
                    if exc.faultCode != 404:
                        raise
                    template = None
                if template is not None:
                    print(f"  {colors.green}  -> template{colors.reset}")
                    lab.device_types[dt].set_template(base, template)

    print(f"{colors.cyan}> workers{colors.reset}")
    if "workers" not in resources:
        print(f"  {colors.yellow}-> SKIP{colors.reset}")
    else:
        if config["version"] >= (2023, 3):
            workers = proxy.scheduler.workers.list(True)
        else:
            workers = proxy.scheduler.workers.list()
        for worker in workers:
            print(f"  {colors.green}* {worker}{colors.reset}")
            data = proxy.scheduler.workers.show(worker)
            lab.workers[worker] = Worker.new(**data)

            if data["default_config"]:
                print(f"  {colors.green}  -> default config{colors.reset}")
            else:
                try:
                    wconfig = str(proxy.scheduler.workers.get_config(worker))
                except xmlrpc.client.Fault as exc:
                    if exc.faultCode != 404:
                        raise
                    wconfig = None
                if wconfig is not None:
                    print(f"  {colors.green}  -> config{colors.reset}")
                    lab.workers[worker].set_config(base, wconfig)

            if data["default_env"]:
                print(f"  {colors.green}  -> default env{colors.reset}")
            else:
                try:
                    wenv = str(proxy.scheduler.workers.get_env(worker))
                except xmlrpc.client.Fault as exc:
                    if exc.faultCode != 404:
                        raise
                    wenv = None
                if wenv is not None:
                    print(f"  {colors.green}  -> env{colors.reset}")
                    lab.workers[worker].set_env(base, wenv)

            if data["default_env_dut"]:
                print(f"  {colors.green}  -> default env-dut{colors.reset}")
            else:
                try:
                    wenv_dut = str(proxy.scheduler.workers.get_env_dut(worker))
                except xmlrpc.client.Fault as exc:
                    if exc.faultCode != 404:
                        raise
                    wenv_dut = None
                if wenv_dut is not None:
                    print(f"  {colors.green}  -> env-dut{colors.reset}")
                    lab.workers[worker].set_env_dut(base, wenv_dut)

    print(f"{colors.cyan}> devices{colors.reset}")
    if "devices" not in resources:
        print(f"  {colors.yellow}-> SKIP{colors.reset}")
    else:
        devices = [d["hostname"] for d in proxy.scheduler.devices.list(True)]
        for device in devices:
            print(f"  {colors.green}* {device}{colors.reset}")
            data = proxy.scheduler.devices.show(device)
            lab.devices[device] = Device.new(**data)
            try:
                ddict = str(proxy.scheduler.devices.get_dictionary(device))
            except xmlrpc.client.Fault as exc:
                if exc.faultCode != 404:
                    raise
                ddict = None
            if ddict is not None:
                print(f"  {colors.green}  -> dictionary{colors.reset}")
                lab.devices[device].set_dict(base, ddict)

    config_file.dump(lab, resources, excluded_resources)

    return 0


def handle_validate(proxy, options, config):
    config_file = ConfigFile(options.config)
    data = config_file.load()
    if validate(data):
        print(f"{colors.green}Config valid.{colors.reset}")
        return 0
    return 1


def handle(proxy, options, config):
    handlers = {
        "apply": handle_apply,
        "import": handle_import,
        "validate": handle_validate,
    }
    return handlers[options.sub_sub_command](proxy, options, config)



# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

from lavacli.utils import VERSION_LATEST, parse_version


def test_parse_version():
    assert parse_version("2018.5.post1-2~bpo9+1.1debian9.1") == (2018, 5)  # nosec
    assert parse_version("2018.5.post1-1+stretch") == (2018, 5)  # nosec
    assert parse_version("2019.1-1-1") == (2019, 1)  # nosec
    assert parse_version("2019.1.post2-1+stretch") == (2019, 1)  # nosec
    assert parse_version("2019.01.post2-1+stretch") == (2019, 1)  # nosec


def test_parse_version_errors():
    assert parse_version(1) == VERSION_LATEST  # nosec
    assert parse_version("201812") == VERSION_LATEST  # nosec



# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import json
import sys
import time
import xmlrpc.client

from lavacli import main


def test_workers_add(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "workers", "add", "worker01"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.add",
                "args": ("worker01", None, False),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_workers_add_1(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "workers",
            "add",
            "worker01",
            "--description",
            "my worker",
            "--disabled",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.add",
                "args": ("worker01", "my worker", True),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_workers_config_delete(setup, monkeypatch, capsys, tmp_path):
    version = "2022.4"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "workers",
            "config",
            "delete",
            "worker01",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.set_config",
                "args": ("worker01", ""),
                "ret": True,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_workers_config_get(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "workers", "config", "get", "worker01"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.get_config",
                "args": ("worker01",),
                "ret": "config content",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "config content\n"  # nosec


def test_workers_config_set(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "config.yaml").open("w") as f_conf:
        f_conf.write("config content")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "workers",
            "config",
            "set",
            "worker01",
            str(tmp_path / "config.yaml"),
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.set_config",
                "args": ("worker01", "config content"),
                "ret": True,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_workers_config_set_error(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "config.yaml").open("w") as f_conf:
        f_conf.write("config content")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "workers",
            "config",
            "set",
            "worker01",
            str(tmp_path / "config.yaml"),
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.set_config",
                "args": ("worker01", "config content"),
                "ret": False,
            },
        ],
    )
    assert main() == 1  # nosec
    assert capsys.readouterr()[1] == "Unable to store worker configuration\n"  # nosec


def test_workers_env_delete(setup, monkeypatch, capsys, tmp_path):
    version = "2022.4"
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "workers", "env", "delete", "worker01"],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.set_env",
                "args": ("worker01", ""),
                "ret": True,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_workers_env_get(setup, monkeypatch, capsys):
    version = "2019.6"
    monkeypatch.setattr(sys, "argv", ["lavacli", "workers", "env", "get", "worker01"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.get_env",
                "args": ("worker01",),
                "ret": "env content",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "env content\n"  # nosec


def test_workers_env_set(setup, monkeypatch, capsys, tmp_path):
    version = "2019.6"
    with (tmp_path / "env.yaml").open("w") as f_conf:
        f_conf.write("env content")
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "workers", "env", "set", "worker01", str(tmp_path / "env.yaml")],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.set_env",
                "args": ("worker01", "env content"),
                "ret": True,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_workers_env_set_error(setup, monkeypatch, capsys, tmp_path):
    version = "2019.6"
    with (tmp_path / "env.yaml").open("w") as f_conf:
        f_conf.write("env content")
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "workers", "env", "set", "worker01", str(tmp_path / "env.yaml")],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.set_env",
                "args": ("worker01", "env content"),
                "ret": False,
            },
        ],
    )
    assert main() == 1  # nosec
    assert capsys.readouterr()[1] == "Unable to store worker environment\n"  # nosec


def test_workers_env_dut_delete(setup, monkeypatch, capsys, tmp_path):
    version = "2022.4"
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "workers", "env-dut", "delete", "worker01"],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.set_env_dut",
                "args": ("worker01", ""),
                "ret": True,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_workers_env_dut_get(setup, monkeypatch, capsys):
    version = "2022.4"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "workers", "env-dut", "get", "worker01"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.get_env_dut",
                "args": ("worker01",),
                "ret": "env-dut content",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "env-dut content\n"  # nosec


def test_workers_env_dut_set(setup, monkeypatch, capsys, tmp_path):
    version = "2022.4"
    with (tmp_path / "env.yaml").open("w") as f_conf:
        f_conf.write("env-dut content")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "workers",
            "env-dut",
            "set",
            "worker01",
            str(tmp_path / "env.yaml"),
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.set_env_dut",
                "args": ("worker01", "env-dut content"),
                "ret": True,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_workers_env_dut_set_error(setup, monkeypatch, capsys, tmp_path):
    version = "2022.4"
    with (tmp_path / "env.yaml").open("w") as f_conf:
        f_conf.write("env-dut content")
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "workers",
            "env-dut",
            "set",
            "worker01",
            str(tmp_path / "env.yaml"),
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.set_env_dut",
                "args": ("worker01", "env-dut content"),
                "ret": False,
            },
        ],
    )
    assert main() == 1  # nosec
    assert capsys.readouterr()[1] == "Unable to store worker dut environment\n"  # nosec


def test_workers_list_exclude_retired(setup, monkeypatch, capsys):
    version = "2023.3"
    monkeypatch.setattr(sys, "argv", ["lavacli", "workers", "list"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.list",
                "args": (False,),
                "ret": ["worker01", "worker02"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Workers:
* worker01
* worker02
"""
    )


def test_workers_list(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "workers", "list"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.list",
                "args": (),
                "ret": ["worker01", "worker02"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Workers:
* worker01
* worker02
"""
    )


def test_workers_list_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "workers", "list", "--json"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.list",
                "args": (),
                "ret": ["worker01", "worker02"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == '["worker01", "worker02"]\n'  # nosec


def test_workers_list_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "workers", "list", "--yaml"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.list",
                "args": (),
                "ret": ["worker01", "worker02"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "[worker01, worker02]\n"  # nosec


def test_workers_maintenance(setup, monkeypatch, capsys):
    version = "2019.1"
    last_ping = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 5  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(sys, "argv", ["lavacli", "workers", "maintenance", "worker01"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.update",
                "args": ("worker01", None, "MAINTENANCE"),
                "ret": None,
            },
            {
                "request": "scheduler.workers.show",
                "args": ("worker01",),
                "ret": {
                    "hostname": "worker01",
                    "description": None,
                    "state": "Idle",
                    "health": "Active",
                    "devices": ["qemu01", "bbb-01"],
                    "last_ping": last_ping,
                },
            },
            {
                "request": "scheduler.devices.list",
                "args": (),
                "ret": [
                    {
                        "hostname": "qemu01",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb-01",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Running",
                        "current_job": 1234,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb-02",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Running",
                        "current_job": 1235,
                        "pipeline": True,
                    },
                ],
            },
            {
                "request": "scheduler.jobs.show",
                "args": (1234,),
                "ret": {
                    "id": "1234",
                    "description": "basic testing",
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Unknown",
                    "state": "Running",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": None,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
            {
                "request": "scheduler.jobs.show",
                "args": (1234,),
                "ret": {
                    "id": "1234",
                    "description": "basic testing",
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": last_ping,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "-> waiting for job 1234\n--> waiting\n"  # nosec


def test_workers_maintenance_nowait(setup, monkeypatch, capsys):
    version = "2019.1"

    def sleep(duration):
        assert duration == 5  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "workers", "maintenance", "worker01", "--no-wait"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.update",
                "args": ("worker01", None, "MAINTENANCE"),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_workers_maintenance_force(setup, monkeypatch, capsys):
    version = "2019.1"
    last_ping = xmlrpc.client.DateTime("20180128T01:01:01")
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "workers", "maintenance", "worker01", "--force"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.update",
                "args": ("worker01", None, "MAINTENANCE"),
                "ret": None,
            },
            {
                "request": "scheduler.workers.show",
                "args": ("worker01",),
                "ret": {
                    "hostname": "worker01",
                    "description": None,
                    "state": "Idle",
                    "health": "Active",
                    "devices": ["qemu01", "bbb-01"],
                    "last_ping": last_ping,
                },
            },
            {
                "request": "scheduler.devices.list",
                "args": (),
                "ret": [
                    {
                        "hostname": "qemu01",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb-01",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Running",
                        "current_job": 1234,
                        "pipeline": True,
                    },
                ],
            },
            {"request": "scheduler.jobs.cancel", "args": (1234,), "ret": None},
            {
                "request": "scheduler.jobs.show",
                "args": (1234,),
                "ret": {
                    "id": "1234",
                    "description": "basic testing",
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Canceled",
                    "state": "Finished",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": last_ping,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "-> waiting for job 1234\n--> canceling\n"  # nosec


def test_workers_show(setup, monkeypatch, capsys):
    version = "2019.1"
    last_ping = xmlrpc.client.DateTime("20180128T01:01:01")
    monkeypatch.setattr(sys, "argv", ["lavacli", "workers", "show", "worker01"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.show",
                "args": ("worker01",),
                "ret": {
                    "hostname": "worker01",
                    "description": None,
                    "state": "Idle",
                    "health": "Active",
                    "devices": ["qemu01", "bbb-01"],
                    "last_ping": last_ping,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """hostname    : worker01
description : None
state       : Idle
health      : Active
devices     : qemu01, bbb-01
last ping   : 20180128T01:01:01
"""
    )


def test_workers_show_json(setup, monkeypatch, capsys):
    version = "2019.1"
    last_ping = xmlrpc.client.DateTime("20180128T01:01:01")
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "workers", "show", "worker01", "--json"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.show",
                "args": ("worker01",),
                "ret": {
                    "hostname": "worker01",
                    "description": None,
                    "state": "Idle",
                    "health": "Active",
                    "devices": ["qemu01", "bbb-01"],
                    "last_ping": last_ping,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == {  # nosec
        "hostname": "worker01",
        "description": None,
        "state": "Idle",
        "health": "Active",
        "devices": ["qemu01", "bbb-01"],
        "last_ping": "20180128T01:01:01",
    }


def test_workers_show_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    last_ping = xmlrpc.client.DateTime("20180128T01:01:01")
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "workers", "show", "worker01", "--yaml"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.show",
                "args": ("worker01",),
                "ret": {
                    "hostname": "worker01",
                    "description": None,
                    "state": "Idle",
                    "health": "Active",
                    "devices": ["qemu01", "bbb-01"],
                    "last_ping": last_ping,
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """description: null
devices: [qemu01, bbb-01]
health: Active
hostname: worker01
last_ping: 20180128T01:01:01
state: Idle
"""
    )


def test_workers_update(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "workers",
            "update",
            "worker01",
            "--description",
            "worker",
            "--health",
            "ACTIVE",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.update",
                "args": ("worker01", "worker", "ACTIVE"),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec



# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import json
import sys
import xmlrpc.client

from lavacli import main


def test_results_job(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "results", "1234"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testjob_results_yaml",
                "args": ("1234",),
                "ret": """[{"suite": "lava", "name": "validate", "result": "pass"},{"suite": "lava", "name": "job", "result": "fail"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Results:
* lava.validate [pass]
* lava.job [fail]
"""
    )


def test_results_job_isatty(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "results", "1234"])
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testjob_results_yaml",
                "args": ("1234",),
                "ret": """[{"suite": "lava", "name": "validate", "result": "pass"}, {"suite": "lava", "name": "boot", "result": "skip"}, {"suite": "lava", "name": "job", "result": "fail"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Results:
* lava.validate [\033[1;32mpass\033[0m]
* lava.boot [skip]
* lava.job [\033[1;31mfail\033[0m]
"""
    )


def test_results_job_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "results", "1234", "--json"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testjob_results_yaml",
                "args": ("1234",),
                "ret": """[{"suite": "lava", "name": "validate", "result": "pass"},{"suite": "lava", "name": "job", "result": "fail"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == [  # nosec
        {"name": "validate", "result": "pass", "suite": "lava"},
        {"name": "job", "result": "fail", "suite": "lava"},
    ]


def test_results_job_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "results", "1234", "--yaml"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testjob_results_yaml",
                "args": ("1234",),
                "ret": """[{"suite": "lava", "name": "validate", "result": "pass"},{"suite": "lava", "name": "job", "result": "fail"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """- {name: validate, result: pass, suite: lava}
- {name: job, result: fail, suite: lava}
"""
    )


def test_results_suite(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "results", "1234", "lava"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testsuite_results_yaml",
                "args": ("1234", "lava"),
                "ret": """[{"suite": "lava", "name": "validate", "result": "pass"},{"suite": "lava", "name": "job", "result": "fail"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Results:
* lava.validate [pass]
* lava.job [fail]
"""
    )


def test_results_suite_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "results", "1234", "lava", "--json"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testsuite_results_yaml",
                "args": ("1234", "lava"),
                "ret": """[{"suite": "lava", "name": "validate", "result": "pass"},{"suite": "lava", "name": "job", "result": "fail"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == [  # nosec
        {"name": "validate", "result": "pass", "suite": "lava"},
        {"name": "job", "result": "fail", "suite": "lava"},
    ]


def test_results_suite_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "results", "1234", "lava", "--yaml"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testsuite_results_yaml",
                "args": ("1234", "lava"),
                "ret": """[{"suite": "lava", "name": "validate", "result": "pass"},{"suite": "lava", "name": "job", "result": "fail"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """- {name: validate, result: pass, suite: lava}
- {name: job, result: fail, suite: lava}
"""
    )


def test_results_case(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "results", "1234", "lava", "validate"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testcase_results_yaml",
                "args": ("1234", "lava", "validate"),
                "ret": """[{"suite": "lava", "name": "validate", "result": "pass"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "pass\n"  # nosec


def test_results_case_isatty_pass(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "results", "1234", "lava", "validate"])
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testcase_results_yaml",
                "args": ("1234", "lava", "validate"),
                "ret": """[{"suite": "lava", "name": "validate", "result": "pass"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "\033[1;32mpass\033[0m\n"  # nosec


def test_results_case_isatty_fail(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "results", "1234", "lava", "validate"])
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testcase_results_yaml",
                "args": ("1234", "lava", "validate"),
                "ret": """[{"suite": "lava", "name": "validate", "result": "fail"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "\033[1;31mfail\033[0m\n"  # nosec


def test_results_case_isatty_skip(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "results", "1234", "lava", "validate"])
    monkeypatch.setattr(sys.stdout, "isatty", lambda: True)
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testcase_results_yaml",
                "args": ("1234", "lava", "validate"),
                "ret": """[{"suite": "lava", "name": "validate", "result": "skip"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "skip\n"  # nosec


def test_results_case_1(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "results", "1234", "lava", "validate"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testcase_results_yaml",
                "args": ("1234", "lava", "validate"),
                "ret": """[{"suite": "lava", "name": "validate", "result": "pass"},{"suite": "lava", "name": "validate", "result": "fail"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "pass\nfail\n"  # nosec


def test_results_case_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "results", "1234", "lava", "validate", "--json"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testcase_results_yaml",
                "args": ("1234", "lava", "validate"),
                "ret": """[{"suite": "lava", "name": "validate", "result": "pass"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == [  # nosec
        {"name": "validate", "result": "pass", "suite": "lava"}
    ]


def test_results_case_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "results", "1234", "lava", "validate", "--yaml"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "results.get_testcase_results_yaml",
                "args": ("1234", "lava", "validate"),
                "ret": """[{"suite": "lava", "name": "validate", "result": "pass"}]""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0] == "- {name: validate, result: pass, suite: lava}\n"
    )



# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import sys
import time
import xmlrpc.client

from lavacli import main


def test_system_active(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "system", "active"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.workers.list",
                "args": (),
                "ret": ["worker01", "worker02"],
            },
            {
                "request": "scheduler.workers.update",
                "args": ("worker01", None, "ACTIVE"),
                "ret": None,
            },
            {
                "request": "scheduler.workers.update",
                "args": ("worker02", None, "ACTIVE"),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Activate workers:
* worker01
* worker02
"""
    )


def test_system_api(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "system", "api"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {"request": "system.api_version", "args": (), "ret": 2},
        ],
    )

    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "2\n"  # nosec


def test_system_export(setup, monkeypatch, capsys, tmp_path):
    last_ping = xmlrpc.client.DateTime("20180128T01:01:01")
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "system", "export", str(tmp_path / "export")]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": "2019.1"},
            {
                "request": "scheduler.aliases.list",
                "args": (),
                "ret": ["alias01", "alias02"],
            },
            {
                "request": "scheduler.aliases.show",
                "args": ("alias01",),
                "ret": {"name": "alias01", "device_types": ["qemu", "kvm"]},
            },
            {
                "request": "scheduler.aliases.show",
                "args": ("alias02",),
                "ret": {"name": "alias02", "device_types": []},
            },
            {
                "request": "scheduler.tags.list",
                "args": (),
                "ret": [
                    {"name": "hdd", "description": "drive attached"},
                    {"name": "tag1", "description": None},
                ],
            },
            {
                "request": "scheduler.workers.list",
                "args": (),
                "ret": ["worker01", "worker02"],
            },
            {
                "request": "scheduler.workers.show",
                "args": ("worker01",),
                "ret": {
                    "hostname": "worker01",
                    "description": None,
                    "state": "Idle",
                    "health": "Active",
                    "devices": ["qemu01", "bbb-01"],
                    "last_ping": last_ping,
                },
            },
            {
                "request": "scheduler.workers.show",
                "args": ("worker02",),
                "ret": {
                    "hostname": "worker01",
                    "description": None,
                    "state": "Idle",
                    "health": "Active",
                    "devices": [],
                    "last_ping": last_ping,
                },
            },
            {
                "request": "scheduler.device_types.list",
                "args": (),
                "ret": [
                    {"name": "bbb", "devices": 1, "installed": True, "template": True},
                    {"name": "qemu", "devices": 1, "installed": True, "template": True},
                ],
            },
            {
                "request": "scheduler.device_types.show",
                "args": ("bbb",),
                "ret": {
                    "name": "bbb",
                    "description": None,
                    "display": True,
                    "owners_only": False,
                    "health_disabled": False,
                    "aliases": [],
                    "devices": ["bbb-01"],
                },
            },
            {
                "request": "scheduler.device_types.get_template",
                "args": ("bbb",),
                "ret": "bbb device-type template",
            },
            {
                "request": "scheduler.device_types.show",
                "args": ("qemu",),
                "ret": {
                    "name": "qemu",
                    "description": None,
                    "display": True,
                    "owners_only": False,
                    "health_disabled": False,
                    "aliases": [],
                    "devices": ["qemu01"],
                },
            },
            {
                "request": "scheduler.device_types.get_template",
                "args": ("qemu",),
                "ret": "qemu device-type template",
            },
            {
                "request": "scheduler.devices.list",
                "args": (False,),
                "ret": [
                    {
                        "hostname": "qemu01",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb-01",
                        "type": "bbb",
                        "health": "Maintenance",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                ],
            },
            {
                "request": "scheduler.devices.show",
                "args": ("qemu01",),
                "ret": {
                    "hostname": "qemu01",
                    "device_type": "qemu",
                    "health": "Good",
                    "state": "Idle",
                    "health_job": True,
                    "description": None,
                    "public": True,
                    "pipeline": True,
                    "has_device_dict": True,
                    "worker": "worker01",
                    "user": None,
                    "group": "group01",
                    "current_job": None,
                    "tags": [],
                },
            },
            {
                "request": "scheduler.devices.get_dictionary",
                "args": ("qemu01",),
                "ret": "qemu01 device dict",
            },
            {
                "request": "scheduler.devices.show",
                "args": ("bbb-01",),
                "ret": {
                    "hostname": "bbb-01",
                    "device_type": "bbb",
                    "health": "Good",
                    "state": "Idle",
                    "health_job": True,
                    "description": None,
                    "public": True,
                    "pipeline": True,
                    "has_device_dict": True,
                    "worker": "worker01",
                    "user": None,
                    "group": "group01",
                    "current_job": None,
                    "tags": [],
                },
            },
            {
                "request": "scheduler.devices.get_dictionary",
                "args": ("bbb-01",),
                "ret": "bbb-01 device dict",
            },
        ],
    )
    assert main() == 0  # nosec
    lines = capsys.readouterr()[0].split("\n")
    assert lines[0] == "Export to %s" % str(tmp_path / "export")  # nosec
    assert (  # nosec
        "\n".join(lines[1:])
        == """Listing aliases
* alias01
* alias02
Listing tags
* hdd
* tag1
Listing workers
* worker01
* worker02
Listing device-types
* bbb
* qemu
Listing devices
* qemu01
* bbb-01
"""
    )


def test_system_maintenance(setup, monkeypatch, capsys):
    last_ping = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 5  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(sys, "argv", ["lavacli", "system", "maintenance"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": "2019.1"},
            {
                "request": "scheduler.workers.list",
                "args": (),
                "ret": ["worker01", "worker02"],
            },
            {
                "request": "scheduler.workers.update",
                "args": ("worker01", None, "MAINTENANCE"),
                "ret": None,
            },
            {
                "request": "scheduler.workers.update",
                "args": ("worker02", None, "MAINTENANCE"),
                "ret": None,
            },
            {
                "request": "scheduler.devices.list",
                "args": (),
                "ret": [
                    {
                        "hostname": "qemu01",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb-01",
                        "type": "bbb",
                        "health": "Good",
                        "state": "Running",
                        "current_job": 1234,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb-02",
                        "type": "bbb",
                        "health": "Good",
                        "state": "Running",
                        "current_job": 1235,
                        "pipeline": True,
                    },
                ],
            },
            {
                "request": "scheduler.jobs.show",
                "args": (1234,),
                "ret": {
                    "id": "1234",
                    "description": "basic testing",
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Unknown",
                    "state": "Running",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": None,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
            {
                "request": "scheduler.jobs.show",
                "args": (1234,),
                "ret": {
                    "id": "1234",
                    "description": "basic testing",
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": last_ping,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
            {
                "request": "scheduler.jobs.show",
                "args": (1235,),
                "ret": {
                    "id": "1235",
                    "description": "basic testing",
                    "device": "bbb-02",
                    "device_type": "bbb",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": last_ping,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Maintenance workers:
* worker01
* worker02
Wait for devices:
* qemu01
* bbb-01
--> waiting for job 1234
---> waiting
* bbb-02
--> waiting for job 1235
"""
    )


def test_system_maintenance_exclude(setup, monkeypatch, capsys):
    last_ping = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 5  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "system", "maintenance", "--exclude", "worker01"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": "2019.1"},
            {
                "request": "scheduler.workers.list",
                "args": (),
                "ret": ["worker01", "worker02"],
            },
            {
                "request": "scheduler.workers.update",
                "args": ("worker02", None, "MAINTENANCE"),
                "ret": None,
            },
            {
                "request": "scheduler.workers.show",
                "args": ("worker01",),
                "ret": {
                    "hostname": "worker01",
                    "description": None,
                    "state": "Idle",
                    "health": "Active",
                    "devices": ["qemu01", "bbb-02"],
                    "last_ping": last_ping,
                },
            },
            {
                "request": "scheduler.devices.list",
                "args": (),
                "ret": [
                    {
                        "hostname": "qemu01",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb-01",
                        "type": "bbb",
                        "health": "Good",
                        "state": "Running",
                        "current_job": 1234,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb-02",
                        "type": "bbb",
                        "health": "Good",
                        "state": "Running",
                        "current_job": 1235,
                        "pipeline": True,
                    },
                ],
            },
            {
                "request": "scheduler.jobs.show",
                "args": (1234,),
                "ret": {
                    "id": "1234",
                    "description": "basic testing",
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Unknown",
                    "state": "Running",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": None,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
            {
                "request": "scheduler.jobs.show",
                "args": (1234,),
                "ret": {
                    "id": "1234",
                    "description": "basic testing",
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": last_ping,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Maintenance workers:
* worker01 [SKIP]
* worker02
Wait for devices:
* bbb-01
--> waiting for job 1234
---> waiting
"""
    )


def test_system_maintenance_force(setup, monkeypatch, capsys):
    last_ping = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 5  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(sys, "argv", ["lavacli", "system", "maintenance", "--force"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": "2019.1"},
            {
                "request": "scheduler.workers.list",
                "args": (),
                "ret": ["worker01", "worker02"],
            },
            {
                "request": "scheduler.workers.update",
                "args": ("worker01", None, "MAINTENANCE"),
                "ret": None,
            },
            {
                "request": "scheduler.workers.update",
                "args": ("worker02", None, "MAINTENANCE"),
                "ret": None,
            },
            {
                "request": "scheduler.devices.list",
                "args": (),
                "ret": [
                    {
                        "hostname": "qemu01",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb-01",
                        "type": "bbb",
                        "health": "Good",
                        "state": "Running",
                        "current_job": 1234,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb-02",
                        "type": "bbb",
                        "health": "Good",
                        "state": "Running",
                        "current_job": 1235,
                        "pipeline": True,
                    },
                ],
            },
            {"request": "scheduler.jobs.cancel", "args": (1234,), "ret": None},
            {
                "request": "scheduler.jobs.show",
                "args": (1234,),
                "ret": {
                    "id": "1234",
                    "description": "basic testing",
                    "device": "bbb-01",
                    "device_type": "bbb",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": last_ping,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
            {"request": "scheduler.jobs.cancel", "args": (1235,), "ret": None},
            {
                "request": "scheduler.jobs.show",
                "args": (1235,),
                "ret": {
                    "id": "1235",
                    "description": "basic testing",
                    "device": "bbb-02",
                    "device_type": "bbb",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": last_ping,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Maintenance workers:
* worker01
* worker02
Wait for devices:
* qemu01
* bbb-01
--> waiting for job 1234
---> canceling
* bbb-02
--> waiting for job 1235
---> canceling
"""
    )


def test_system_methods_list(setup, monkeypatch, capsys):
    monkeypatch.setattr(sys, "argv", ["lavacli", "system", "methods", "list"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": "2019.1"},
            {
                "request": "system.listMethods",
                "args": (),
                "ret": [
                    "scheduler.job_details",
                    "scheduler.job_health",
                    "scheduler.job_list_status",
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == "scheduler.job_details\nscheduler.job_health\nscheduler.job_list_status\n"
    )


def test_system_methods_help(setup, monkeypatch, capsys):
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "system", "methods", "help", "system.version"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": "2019.1"},
            {
                "request": "system.methodHelp",
                "args": ("system.version",),
                "ret": """Name
----
`system.version` ()

Description
-----------
Return the lava-server version string

Arguments
---------
None

Return value
------------
lava-server version string""",
            },
        ],
    )

    assert main() == 0  # nosec

    assert (  # nosec
        capsys.readouterr()[0]
        == """Name
----
`system.version` ()

Description
-----------
Return the lava-server version string

Arguments
---------
None

Return value
------------
lava-server version string
"""
    )


def test_system_methods_signature(setup, monkeypatch, capsys):
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "system", "methods", "signature", "system.version"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": "2019.1"},
            {
                "request": "system.methodSignature",
                "args": ("system.version",),
                "ret": "undef",
            },
        ],
    )

    assert main() == 0  # nosec

    assert capsys.readouterr()[0] == "undef\n"  # nosec


def test_system_version(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "system", "version"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {"request": "system.version", "args": (), "ret": version},
        ],
    )

    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "2019.1\n"  # nosec


def test_system_whoami(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "system", "whoami"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {"request": "system.whoami", "args": (), "ret": "lava-admin"},
        ],
    )

    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "lava-admin\n"  # nosec


def test_system_whoami_anonymous(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "system", "whoami"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {"request": "system.whoami", "args": (), "ret": None},
        ],
    )

    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "<AnonymousUser>\n"  # nosec



# vim: set ts=4

# Copyright 2022-present Linaro
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import pytest
from voluptuous import MultipleInvalid, Schema

from lavacli import schemas
from lavacli.utils import safe_yaml


def test_user_group_must_exist():
    config = """groups:
  group1:
users:
  user1:
    groups:
    - group2
"""
    schema = Schema(schemas.user_group_must_exist)
    with pytest.raises(MultipleInvalid) as exc:
        schema(safe_yaml.load(config))
    assert exc.value.msg == "Group must exist"
    assert exc.value.path == ["users", "user1", "groups", "group2"]


def test_device_dt_must_exist():
    config = """device_types:
  qemu:
devices:
  docker-01:
    device_type: docker
"""
    schema = Schema(schemas.device_dt_must_exist)
    with pytest.raises(MultipleInvalid) as exc:
        schema(safe_yaml.load(config))
    assert exc.value.msg == "Device type must exist"
    assert exc.value.path == ["devices", "docker-01", "device_type", "docker"]


def test_device_worker_must_exist():
    config = """device_types:
  qemu:
devices:
  docker-01:
    worker: worker01
workers:
  worker02:
"""
    schema = Schema(schemas.device_worker_must_exist)
    with pytest.raises(MultipleInvalid) as exc:
        schema(safe_yaml.load(config))
    assert exc.value.msg == "Worker must exist"
    assert exc.value.path == ["devices", "docker-01", "worker", "worker01"]


def test_device_perms_group_must_exist():
    config = """devices:
  docker-01:
    permissions:
    - name: change_device
      group: group2
groups:
  group1:
"""
    schema = Schema(schemas.device_perms_group_must_exist)
    with pytest.raises(MultipleInvalid) as exc:
        schema(safe_yaml.load(config))
    assert exc.value.msg == "Group must exist"
    assert exc.value.path == [
        "devices",
        "docker-01",
        "permissions",
        {"name": "change_device", "group": "group2"},
    ]


def test_dt_perms_group_must_exist():
    config = """device_types:
  docker:
    permissions:
    - name: change_devicetype
      group: group2
groups:
  group1:
"""
    schema = Schema(schemas.dt_perms_group_must_exist)
    with pytest.raises(MultipleInvalid) as exc:
        schema(safe_yaml.load(config))
    assert exc.value.msg == "Group must exist"
    assert exc.value.path == [
        "devices_types",
        "docker",
        "permissions",
        {"name": "change_devicetype", "group": "group2"},
    ]



# vim: set ts=4

# Copyright 2018 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import json
import sys
import time
import xmlrpc.client

from lavacli import main


def test_devices_add(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "devices",
            "add",
            "qemu01",
            "--type",
            "qemu",
            "--worker",
            "worker01",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.add",
                "args": ("qemu01", "qemu", "worker01", None, None, True, None, None),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_devices_add_1(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "devices",
            "add",
            "qemu01",
            "--type",
            "qemu",
            "--worker",
            "worker01",
            "--health",
            "GOOD",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.add",
                "args": ("qemu01", "qemu", "worker01", None, None, True, "GOOD", None),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_devices_add_2(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "devices",
            "add",
            "qemu01",
            "--type",
            "qemu",
            "--worker",
            "worker01",
            "--user",
            "me",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.add",
                "args": ("qemu01", "qemu", "worker01", "me", None, True, None, None),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_devices_dict_delete(setup, monkeypatch, capsys):
    version = "2022.4"
    monkeypatch.setattr(sys, "argv", ["lavacli", "devices", "dict", "delete", "qemu01"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.set_dictionary",
                "args": ("qemu01", ""),
                "ret": True,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_devices_dict_get(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "devices", "dict", "get", "qemu01"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.get_dictionary",
                "args": ("qemu01", False, None),
                "ret": "yaml_dict",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "yaml_dict\n"  # nosec


def test_devices_dict_get_render_field(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "devices", "dict", "get", "--render", "qemu01", "hello.0.world"],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.get_dictionary",
                "args": ("qemu01", True, None),
                "ret": """{"hello": [{"world": "as usual"}, "my"]}""",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "as usual\n"  # nosec


def test_devices_dict_get_render_field_out_of_range(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "devices", "dict", "get", "--render", "qemu01", "hello.2.world"],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.get_dictionary",
                "args": ("qemu01", True, None),
                "ret": """{"hello": [{"world": "as usual"}, "my"]}""",
            },
        ],
    )
    assert main() == 1  # nosec
    assert capsys.readouterr()[1] == "list index out of range (2 vs 2)\n"  # nosec


def test_devices_dict_get_render_field_missing(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "devices", "dict", "get", "--render", "qemu01", "hello.0.worl"],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.get_dictionary",
                "args": ("qemu01", True, None),
                "ret": """{"hello": [{"world": "as usual"}, "my"]}""",
            },
        ],
    )
    assert main() == 1  # nosec
    assert (  # nosec
        capsys.readouterr()[1] == "Unknown key 'worl' for '{'world': 'as usual'}'\n"
    )


def test_devices_dict_get_render_field_missing_1(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "devices", "dict", "get", "--render", "qemu01", "hello.0.world.0"],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.get_dictionary",
                "args": ("qemu01", True, None),
                "ret": """{"hello": [{"world": "as usual"}, "my"]}""",
            },
        ],
    )
    assert main() == 1  # nosec
    assert (  # nosec
        capsys.readouterr()[1] == "Unable to lookup inside 'as usual' for '0'\n"
    )


def test_devices_dict_get_render_field_missing_2(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "devices",
            "dict",
            "get",
            "--render",
            "qemu01",
            "hello.0.world.missing",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.get_dictionary",
                "args": ("qemu01", True, None),
                "ret": """{"hello": [{"world": "as usual"}, "my"]}""",
            },
        ],
    )
    assert main() == 1  # nosec
    assert (  # nosec
        capsys.readouterr()[1] == "Unable to lookup inside 'as usual' for 'missing'\n"
    )


def test_devices_dict_get_jinja2_field(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "devices", "dict", "get", "qemu01", "hello"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.get_dictionary",
                "args": ("qemu01", False, None),
                "ret": "{% set hello = 'bla' %}",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "bla\n"  # nosec


def test_devices_dict_get_jinja2_field_2(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "devices", "dict", "get", "qemu01", "hello.0.bla"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.get_dictionary",
                "args": ("qemu01", False, None),
                "ret": "{% set hello = [{'bla': 'something'}] %}",
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "something\n"  # nosec


def test_devices_dict_get_jinja2_field_missing(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "devices", "dict", "get", "qemu01", "world"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.get_dictionary",
                "args": ("qemu01", False, None),
                "ret": "{% set hello = 'bla' %}",
            },
        ],
    )
    assert main() == 1  # nosec
    assert capsys.readouterr()[1] == "Unknown field 'world'\n"  # nosec


def test_devices_dict_set(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "dict.jinja2").open("w") as f_conf:
        f_conf.write("{% set exclusive = True %}")
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "devices", "dict", "set", "qemu01", str(tmp_path / "dict.jinja2")],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.set_dictionary",
                "args": ("qemu01", "{% set exclusive = True %}"),
                "ret": True,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_devices_dict_set_error(setup, monkeypatch, capsys, tmp_path):
    version = "2019.1"
    with (tmp_path / "dict.jinja2").open("w") as f_conf:
        f_conf.write("{% set exclusive = True %}")
    monkeypatch.setattr(
        sys,
        "argv",
        ["lavacli", "devices", "dict", "set", "qemu01", str(tmp_path / "dict.jinja2")],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.set_dictionary",
                "args": ("qemu01", "{% set exclusive = True %}"),
                "ret": False,
            },
        ],
    )
    assert main() == 1  # nosec
    assert capsys.readouterr()[1] == "Unable to set the configuration\n"  # nosec


def test_devices_list(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "devices", "list"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.list",
                "args": (False,),
                "ret": [
                    {
                        "hostname": "qemu01",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                    {
                        "hostname": "qemu02",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Running",
                        "current_job": 1234,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb01",
                        "type": "bbb",
                        "health": "Maintenance",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """Devices:
* qemu01 (qemu): Idle,Good
* qemu02 (qemu): Running,Good
* bbb01 (bbb): Idle,Maintenance
"""
    )


def test_devices_list_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "devices", "list", "--json"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.list",
                "args": (False,),
                "ret": [
                    {
                        "hostname": "qemu01",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                    {
                        "hostname": "qemu02",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Running",
                        "current_job": 1234,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb01",
                        "type": "bbb",
                        "health": "Maintenance",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == [  # nosec
        {
            "hostname": "qemu01",
            "type": "qemu",
            "health": "Good",
            "state": "Idle",
            "current_job": None,
            "pipeline": True,
        },
        {
            "hostname": "qemu02",
            "type": "qemu",
            "health": "Good",
            "state": "Running",
            "current_job": 1234,
            "pipeline": True,
        },
        {
            "hostname": "bbb01",
            "type": "bbb",
            "health": "Maintenance",
            "state": "Idle",
            "current_job": None,
            "pipeline": True,
        },
    ]


def test_devices_list_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "devices", "list", "--yaml"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.list",
                "args": (False,),
                "ret": [
                    {
                        "hostname": "qemu01",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                    {
                        "hostname": "qemu02",
                        "type": "qemu",
                        "health": "Good",
                        "state": "Running",
                        "current_job": 1234,
                        "pipeline": True,
                    },
                    {
                        "hostname": "bbb01",
                        "type": "bbb",
                        "health": "Maintenance",
                        "state": "Idle",
                        "current_job": None,
                        "pipeline": True,
                    },
                ],
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """- {current_job: null, health: Good, hostname: qemu01, pipeline: true, state: Idle,
  type: qemu}
- {current_job: 1234, health: Good, hostname: qemu02, pipeline: true, state: Running,
  type: qemu}
- {current_job: null, health: Maintenance, hostname: bbb01, pipeline: true, state: Idle,
  type: bbb}
"""
    )


def test_devices_maintenance(setup, monkeypatch, capsys):
    version = "2019.1"
    last_ping = xmlrpc.client.DateTime("20180128T01:01:01")

    def sleep(duration):
        assert duration == 5  # nosec

    monkeypatch.setattr(time, "sleep", sleep)
    monkeypatch.setattr(sys, "argv", ["lavacli", "devices", "maintenance", "qemu01"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.update",
                "args": ("qemu01", None, None, None, None, "MAINTENANCE", None),
                "ret": None,
            },
            {
                "request": "scheduler.devices.show",
                "args": ("qemu01",),
                "ret": {
                    "hostname": "qemu01",
                    "device_type": "qemu",
                    "health": "Good",
                    "state": "Idle",
                    "health_job": True,
                    "description": None,
                    "public": True,
                    "pipeline": True,
                    "has_device_dict": True,
                    "worker": "worker01",
                    "user": None,
                    "group": "group01",
                    "current_job": 1234,
                    "tags": ["a", "b"],
                },
            },
            {
                "request": "scheduler.jobs.show",
                "args": (1234,),
                "ret": {
                    "id": "1234",
                    "description": "basic testing",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Unknown",
                    "state": "Running",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": None,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
            {
                "request": "scheduler.jobs.show",
                "args": (1234,),
                "ret": {
                    "id": "1234",
                    "description": "basic testing",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": last_ping,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "-> waiting for job 1234\n--> waiting\n"  # nosec


def test_devices_maintenance_force(setup, monkeypatch, capsys):
    version = "2019.1"
    last_ping = xmlrpc.client.DateTime("20180128T01:01:01")
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "devices", "maintenance", "qemu01", "--force"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.update",
                "args": ("qemu01", None, None, None, None, "MAINTENANCE", None),
                "ret": None,
            },
            {
                "request": "scheduler.devices.show",
                "args": ("qemu01",),
                "ret": {
                    "hostname": "qemu01",
                    "device_type": "qemu",
                    "health": "Good",
                    "state": "Idle",
                    "health_job": True,
                    "description": None,
                    "public": True,
                    "pipeline": True,
                    "has_device_dict": True,
                    "worker": "worker01",
                    "user": None,
                    "group": "group01",
                    "current_job": 1234,
                    "tags": ["a", "b"],
                },
            },
            {"request": "scheduler.jobs.cancel", "args": (1234,), "ret": None},
            {
                "request": "scheduler.jobs.show",
                "args": (1234,),
                "ret": {
                    "id": "1234",
                    "description": "basic testing",
                    "device": "qemu01",
                    "device_type": "qemu",
                    "health_check": False,
                    "pipeline": True,
                    "health": "Complete",
                    "state": "Finished",
                    "submitter": "lava-bot",
                    "submit_time": last_ping,
                    "start_time": last_ping,
                    "end_time": last_ping,
                    "tags": [],
                    "visibility": "Publicly visible",
                    "failure_comment": "",
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "-> waiting for job 1234\n--> canceling\n"  # nosec


def test_devices_maintenance_without_current_job(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "devices", "maintenance", "qemu01", "--force"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.update",
                "args": ("qemu01", None, None, None, None, "MAINTENANCE", None),
                "ret": None,
            },
            {
                "request": "scheduler.devices.show",
                "args": ("qemu01",),
                "ret": {
                    "hostname": "qemu01",
                    "device_type": "qemu",
                    "health": "Good",
                    "state": "Idle",
                    "health_job": True,
                    "description": None,
                    "public": True,
                    "pipeline": True,
                    "has_device_dict": True,
                    "worker": "worker01",
                    "user": None,
                    "group": "group01",
                    "current_job": None,
                    "tags": ["a", "b"],
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_devices_show(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "devices", "show", "qemu01"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.show",
                "args": ("qemu01",),
                "ret": {
                    "hostname": "qemu01",
                    "device_type": "qemu",
                    "health": "Good",
                    "state": "Idle",
                    "health_job": True,
                    "description": None,
                    "public": True,
                    "pipeline": True,
                    "has_device_dict": True,
                    "worker": "worker01",
                    "user": None,
                    "group": "group01",
                    "current_job": None,
                    "tags": ["a", "b"],
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """name        : qemu01
device-type : qemu
state       : Idle
health      : Good
user        : None
group       : group01
health job  : True
description : None
public      : True
pipeline    : True
device-dict : True
worker      : worker01
current job : None
tags        : ['a', 'b']
"""
    )


def test_devices_show_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "devices", "show", "qemu01", "--json"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.show",
                "args": ("qemu01",),
                "ret": {
                    "hostname": "qemu01",
                    "device_type": "qemu",
                    "health": "Good",
                    "state": "Idle",
                    "health_job": True,
                    "description": None,
                    "public": True,
                    "pipeline": True,
                    "has_device_dict": True,
                    "worker": "worker01",
                    "user": None,
                    "group": "group01",
                    "current_job": None,
                    "tags": ["a", "b"],
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert json.loads(capsys.readouterr()[0]) == {  # nosec
        "hostname": "qemu01",
        "device_type": "qemu",
        "health": "Good",
        "state": "Idle",
        "health_job": True,
        "description": None,
        "public": True,
        "pipeline": True,
        "has_device_dict": True,
        "worker": "worker01",
        "user": None,
        "group": "group01",
        "current_job": None,
        "tags": ["a", "b"],
    }


def test_devices_show_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "devices", "show", "qemu01", "--yaml"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.show",
                "args": ("qemu01",),
                "ret": {
                    "hostname": "qemu01",
                    "device_type": "qemu",
                    "health": "Good",
                    "state": "Idle",
                    "health_job": True,
                    "description": None,
                    "public": True,
                    "pipeline": True,
                    "has_device_dict": True,
                    "worker": "worker01",
                    "user": None,
                    "group": "group01",
                    "current_job": None,
                    "tags": ["a", "b"],
                },
            },
        ],
    )
    assert main() == 0  # nosec
    assert (  # nosec
        capsys.readouterr()[0]
        == """current_job: null
description: null
device_type: qemu
group: group01
has_device_dict: true
health: Good
health_job: true
hostname: qemu01
pipeline: true
public: true
state: Idle
tags: [a, b]
user: null
worker: worker01
"""
    )


def test_devices_tags_add(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "devices", "tags", "add", "qemu01", "hdd"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.tags.add",
                "args": ("qemu01", "hdd"),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_devices_tags_delete(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "devices", "tags", "delete", "qemu01", "hdd"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.tags.delete",
                "args": ("qemu01", "hdd"),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec


def test_devices_tags_list(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(sys, "argv", ["lavacli", "devices", "tags", "list", "qemu01"])
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.tags.list",
                "args": ("qemu01",),
                "ret": ["hdd", "virt"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "Tags:\n* hdd\n* virt\n"  # nosec


def test_devices_tags_list_json(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "devices", "tags", "list", "--json", "qemu01"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.tags.list",
                "args": ("qemu01",),
                "ret": ["hdd", "virt"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == '["hdd", "virt"]\n'  # nosec


def test_devices_tags_list_yaml(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys, "argv", ["lavacli", "devices", "tags", "list", "--yaml", "qemu01"]
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.tags.list",
                "args": ("qemu01",),
                "ret": ["hdd", "virt"],
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == "[hdd, virt]\n"  # nosec


def test_devices_update(setup, monkeypatch, capsys):
    version = "2019.1"
    monkeypatch.setattr(
        sys,
        "argv",
        [
            "lavacli",
            "devices",
            "update",
            "qemu01",
            "--worker",
            "worker01",
            "--health",
            "UNKNOWN",
        ],
    )
    monkeypatch.setattr(
        xmlrpc.client.ServerProxy,
        "data",
        [
            {"request": "system.version", "args": (), "ret": version},
            {
                "request": "scheduler.devices.update",
                "args": ("qemu01", "worker01", None, None, None, "UNKNOWN", None),
                "ret": None,
            },
        ],
    )
    assert main() == 0  # nosec
    assert capsys.readouterr()[0] == ""  # nosec



# vim: set ts=4

# Copyright 2017 Rémi Duraffort
# This file is part of lavacli.
#
# lavacli is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# lavacli is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with lavacli.  If not, see <http://www.gnu.org/licenses/>

import xmlrpc.client

import pytest

from lavacli.utils import safe_yaml


class RecordingProxyFactory:
    def __new__(cls, proxy_data):
        class RecordingProxy:
            data = proxy_data

            def __init__(self, uri, allow_none, transport):
                self.request = []

            def __call__(self, *args):
                request = ".".join(self.request)
                self.request = []
                data = self.data.pop(0)
                assert request == data["request"]  # nosec
                assert args == data["args"]  # nosec
                ret = data["ret"]
                if isinstance(ret, dict) and ret.get("fault_code"):
                    raise xmlrpc.client.Fault(ret.get("fault_code"), "fault string")
                return ret

            def __getattr__(self, attr):
                self.request.append(attr)
                return self

        return RecordingProxy


@pytest.fixture
def setup(monkeypatch, tmp_path):
    monkeypatch.setenv("XDG_CONFIG_HOME", str(tmp_path))
    with (tmp_path / "lavacli.yaml").open("w") as f_conf:
        safe_yaml.dump({"default": {"uri": "https://lava.example.com/RPC2"}}, f_conf)
    monkeypatch.setattr(xmlrpc.client, "ServerProxy", RecordingProxyFactory(None))
